@@ -2,11 +2,11 @@
 **  Copyright (c) 2004-2007 Sendmail, Inc. and its suppliers.
 **	All rights reserved.
 **
-**  $Id: dk-filter.c,v 1.162 2007/03/13 18:10:40 msk Exp $
+**  $Id: dk-filter.c,v 1.169 2007/04/08 00:22:44 msk Exp $
 */
 
 #ifndef lint
-static char dk_filter_c_id[] = "@(#)$Id: dk-filter.c,v 1.162 2007/03/13 18:10:40 msk Exp $";
+static char dk_filter_c_id[] = "@(#)$Id: dk-filter.c,v 1.169 2007/04/08 00:22:44 msk Exp $";
 #endif /* !lint */
 
 /* system includes */
@@ -167,6 +167,7 @@ struct msgctx
 typedef struct connctx * connctx;
 struct connctx
 {
+	bool		cctx_noleadspc;		/* no leading spaces */
 	char		cctx_host[MAXHOSTNAMELEN + 1];
 						/* hostname */
 	_SOCK_ADDR	cctx_ip;		/* IP info */
@@ -289,6 +290,9 @@ static void dkf_cleanup __P((SMFICTX *));
 bool addxhdr;					/* add identifying header? */
 bool hdrlist;					/* signature includes hdrlist */
 bool dolog;					/* syslog */
+#if _FFR_QUARANTINE
+bool quarantine;				/* quarantine failures? */
+#endif /* _FFR_QUARANTINE */
 bool no_i_whine;				/* noted ${i} is undefined */
 #if _FFR_REPORTINFO
 bool send_reports;				/* send failure reports */
@@ -784,11 +788,17 @@ dkf_libstatus(SMFICTX *ctx, char *where, int status)
 		retcode = config.cfg_internal;
 		if (dolog)
 		{
+			const char *err;
+
+			err = dk_geterror(dfc->mctx_dk);
+			if (err == NULL)
+				err = strerror(errno);
+
 			syslog(LOG_ERR,
 			       "%s: %s%sinternal error from libdk: %s",
 			       JOBID(dfc->mctx_jobid),
 			       where == NULL ? "" : where,
-			       where == NULL ? "" : ": ", strerror(errno));
+			       where == NULL ? "" : ": ", err);
 		}
 		break;
 
@@ -814,10 +824,16 @@ dkf_libstatus(SMFICTX *ctx, char *where, int status)
 		retcode = config.cfg_internal;
 		if (dolog)
 		{
+			const char *err;
+
+			err = dk_geterror(dfc->mctx_dk);
+			if (err == NULL)
+				err = strerror(errno);
+
 			syslog(LOG_ERR, "%s: %s%sresource unavailable: %s",
 			       JOBID(dfc->mctx_jobid),
 			       where == NULL ? "" : where,
-			       where == NULL ? "" : ": ", strerror(errno));
+			       where == NULL ? "" : ": ", err);
 		}
 		break;
 
@@ -825,8 +841,15 @@ dkf_libstatus(SMFICTX *ctx, char *where, int status)
 		retcode = config.cfg_badsig;
 		if (dolog)
 		{
-			syslog(LOG_ERR, "%s: signature verification failed",
-			       JOBID(dfc->mctx_jobid));
+			const char *err;
+
+			err = dk_geterror(dfc->mctx_dk);
+			if (err == NULL)
+				err = "unknown cause";
+
+			syslog(LOG_ERR,
+			       "%s: signature verification failed: %s",
+			       JOBID(dfc->mctx_jobid), err);
 		}
 		break;
 
@@ -834,8 +857,14 @@ dkf_libstatus(SMFICTX *ctx, char *where, int status)
 		retcode = config.cfg_badsig;
 		if (dolog)
 		{
-			syslog(LOG_ERR, "%s: syntax error in signature data",
-			       JOBID(dfc->mctx_jobid));
+			const char *err;
+
+			err = dk_geterror(dfc->mctx_dk);
+			if (err == NULL)
+				err = "unspecified";
+
+			syslog(LOG_ERR, "%s: syntax error: %s",
+			       JOBID(dfc->mctx_jobid), err);
 		}
 		break;
 	}
@@ -1313,6 +1342,65 @@ dkf_report(msgctx dfc, char *result, char *reason)
 **  BEGIN milter section
 */
 
+#if SMFI_VERSION >= 0x01000000
+/*
+**  MLFI_NEGOTIATE -- handler called on new SMTP connection to negotiate
+**                    MTA options
+**
+**  Parameters:
+**  	ctx -- milter context
+**	f0  -- actions offered by the MTA
+**	f1  -- protocol steps offered by the MTA
+**	f2  -- reserved for future extensions
+**	f3  -- reserved for future extensions
+**	pf0 -- actions requested by the milter
+**	pf1 -- protocol steps requested by the milter
+**	pf2 -- reserved for future extensions
+**	pf3 -- reserved for future extensions
+**
+**  Return value:
+**  	An SMFIS_* constant.
+*/
+
+static sfsistat
+mlfi_negotiate(SMFICTX *ctx,
+	unsigned long f0, unsigned long f1,
+	unsigned long f2 __attribute__((unused)),
+	unsigned long f3 __attribute__((unused)),
+	unsigned long *pf0, unsigned long *pf1,
+	unsigned long *pf2, unsigned long *pf3)
+{
+	connctx cc;
+
+#ifdef SMFIF_QUARANTINE
+	*pf0 = SMFIF_ADDHDRS|SMFIF_CHGHDRS|SMFIF_QUARANTINE;
+#else /* SMFIF_QUARANTINE */
+	*pf0 = SMFIF_ADDHDRS|SMFIF_CHGHDRS;
+#endif /* SMFIF_QUARANTINE */
+
+	*pf1 = 0;
+#ifdef SMFIP_HDR_LEADSPC
+	if ((f1 & SMFIP_HDR_LEADSPC) != 0)
+	{
+		cc = malloc(sizeof(struct connctx));
+		if (cc != NULL)
+		{
+			memset(cc, '\0', sizeof(struct connctx));
+			cc->cctx_noleadspc = TRUE;
+			*pf1 |= SMFIP_HDR_LEADSPC;
+
+			smfi_setpriv(ctx, cc);
+		}
+	}
+#endif /* SMFIP_HDR_LEADSPC */
+
+	*pf2 = 0;
+	*pf3 = 0;
+
+	return SMFIS_CONTINUE;
+}
+#endif /* SMFI_VERSION >= 0x01000000 */
+
 /*
 **  MLFI_CONNECT -- connection handler
 **
@@ -1359,16 +1447,22 @@ mlfi_connect(SMFICTX *ctx, char *host, _SOCK_ADDR *ip)
 	}
 
 	/* copy hostname and IP information to a connection context */
-	cc = malloc(sizeof(struct connctx));
+	cc = smfi_getpriv(ctx);
 	if (cc == NULL)
 	{
-		if (dolog)
+		cc = malloc(sizeof(struct connctx));
+		if (cc == NULL)
 		{
-			syslog(LOG_ERR, "%s malloc(): %s", host,
-			       strerror(errno));
+			if (dolog)
+			{
+				syslog(LOG_ERR, "%s malloc(): %s", host,
+				       strerror(errno));
+			}
+
+			return SMFIS_TEMPFAIL;
 		}
 
-		return SMFIS_TEMPFAIL;
+		memset(cc, '\0', sizeof(struct connctx));
 	}
 
 	sm_strlcpy(cc->cctx_host, host, sizeof cc->cctx_host);
@@ -1474,6 +1568,12 @@ mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
 	msgctx dfc;
 	connctx cc;
 	Header newhdr;
+#ifdef _FFR_ANTICIPATE_SENDMAIL_MUNGE
+	char *end;
+	char *p;
+	char *q;
+	char hv[MAXHEADER];
+#endif /* _FFR_ANTICIPATE_SENDMAIL_MUNGE */
 
 #ifndef DEBUG
 	assert(ctx != NULL);
@@ -1501,7 +1601,41 @@ mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
 
 	(void) memset(newhdr, '\0', sizeof(struct Header));
 	newhdr->hdr_hdr = strdup(headerf);
+
+#ifdef _FFR_ANTICIPATE_SENDMAIL_MUNGE
+	/*
+	**  The sendmail MTA does some minor header rewriting on outgoing
+	**  mail.  This makes things slightly prettier for the MUA, but
+	**  these changes are made after this filter has already generated
+	**  and added a signature.  As a result, verification of the
+	**  signature will fail because what got signed isn't the same
+	**  as what actually goes out.  This chunk of code attempts to
+	**  compensate by arranging to feed to the canonicalization
+	**  algorithms the headers exactly as the MTA will modify them, so
+	**  verification should still work.  This is based on experimentation
+	**  and on reading sendmail/headers.c, and may require more tweaking
+	**  before it's precisely right.
+	*/
+
+	(void) memset(hv, '\0', sizeof hv);
+	
+	end = hv + sizeof hv;
+
+	for (p = headerv, q = hv; *p != '\0' && q < end; p++)
+	{
+		/* skip initial spaces */
+		if (q == hv && isascii(*p) && isspace(*p))
+			continue;
+
+		*q = *p;
+		q++;
+	}
+
+	newhdr->hdr_val = strdup(hv);
+#else /* _FFR_ANTICIPATE_SENDMAIL_MUNGE */
 	newhdr->hdr_val = strdup(headerv);
+#endif /* _FFR_ANTICIPATE_SENDMAIL_MUNGE */
+
 	newhdr->hdr_next = NULL;
 
 	if (newhdr->hdr_hdr == NULL || newhdr->hdr_val == NULL)
@@ -1932,7 +2066,8 @@ mlfi_eoh(SMFICTX *ctx)
 				continue;
 		}
 
-		snprintf(hdrbuf, MAXHEADER, "%s: %s%s", hdr->hdr_hdr,
+		snprintf(hdrbuf, MAXHEADER, "%s:%s%s%s", hdr->hdr_hdr,
+		         cc->cctx_noleadspc ? "" : " ",
 		         hdr->hdr_val, CRLF);
 		status = dk_header(dfc->mctx_dk, hdrbuf, strlen(hdrbuf));
 		if (status != DK_STAT_OK)
@@ -2157,6 +2292,35 @@ mlfi_eom(SMFICTX *ctx)
 			return dkf_libstatus(ctx, "dk_eom()", status);
 		}
 
+#ifdef _FFR_QUARANTINE
+# ifdef SMFIF_QUARANTINE
+		/* quarantine for "bad" results if requested */
+		if (quarantine &&
+		    (status == DK_STAT_BADSIG ||
+		     (status == DK_STAT_NOSIG && dfc->mctx_addheader)))
+		{
+			char *failstatus;
+			char qreason[BUFRSZ + 1];
+
+			failstatus = "bad signature";
+			if (status == DK_STAT_NOSIG)
+				failstatus = "no signature";
+
+			snprintf(qreason, sizeof qreason, "%s: %s",
+			         progname, failstatus);
+			if (smfi_quarantine(ctx, qreason) != MI_SUCCESS)
+			{
+				if (dolog)
+				{
+					syslog(LOG_ERR,
+					       "%s smfi_quarantine() failed",
+					       dfc->mctx_jobid);
+				}
+			}
+		}
+# endif /* SMFIF_QUARANTINE */
+#endif /* _FFR_QUARANTINE */
+
 		/* compute and insert the signature, if we're signing */
 		if (dfc->mctx_signing)
 		{
@@ -2219,7 +2383,8 @@ mlfi_eom(SMFICTX *ctx)
 
 			memset(header, '\0', sizeof header);
 			snprintf(header, sizeof header,
-			         "a=%s; s=%s; d=%s; c=%s; q=%s;\n\t%sb=",
+			         "%sa=%s; s=%s; d=%s; c=%s; q=%s;\n\t%sb=",
+			         cc->cctx_noleadspc ? " " : "",
 			         dkf_signalg(dfc->mctx_signalg),
 #if _FFR_MULTIPLE_KEYS
 			         sel == NULL ? DK_DEFAULT_SELECTOR : sel,
@@ -2306,7 +2471,8 @@ mlfi_eom(SMFICTX *ctx)
 			                      val, sizeof val);
 
 			snprintf(header, sizeof header,
-			         "%s %s=%s; domainkeys=%s%s%s%s%s",
+			         "%s%s %s=%s; domainkeys=%s%s%s%s%s",
+			         cc->cctx_noleadspc ? " " : "",
 			         hostname, hdr, val, authresult,
 			         comment == NULL ? "" : " (",
 			         comment == NULL ? "" : comment,
@@ -2316,7 +2482,8 @@ mlfi_eom(SMFICTX *ctx)
 		else
 		{
 			snprintf(header, sizeof header,
-			         "%s; domainkeys=%s%s%s%s%s",
+			         "%s%s; domainkeys=%s%s%s%s%s",
+			         cc->cctx_noleadspc ? " " : "",
 			         hostname, authresult,
 			         comment == NULL ? "" : " (",
 			         comment == NULL ? "" : comment,
@@ -2352,7 +2519,8 @@ mlfi_eom(SMFICTX *ctx)
 
 		memset(xfhdr, '\0', sizeof xfhdr);
 
-		snprintf(xfhdr, MAXHEADER, "%s v%s %s %s", DKF_PRODUCT,
+		snprintf(xfhdr, sizeof xfhdr, "%s%s v%s %s %s", DKF_PRODUCT,
+		         cc->cctx_noleadspc ? " " : "",
 		         DKF_VERSION, hostname,
 		         dfc->mctx_jobid != NULL ? dfc->mctx_jobid
 		                                : JOBIDUNKNOWN);
@@ -2482,7 +2650,11 @@ struct smfiDesc smfilter =
 {
 	DKF_PRODUCT,	/* filter name */
 	SMFI_VERSION,	/* version code -- do not change */
+#ifdef SMFIF_QUARANTINE
+	(SMFIF_ADDHDRS|SMFIF_CHGHDRS|SMFIF_QUARANTINE),	/* flags */
+#else /* SMFIF_QUARANTINE */
 	(SMFIF_ADDHDRS|SMFIF_CHGHDRS),	/* flags */
+#endif /* SMFIF_QUARANTINE */
 	mlfi_connect,	/* connection info filter */
 	NULL,		/* SMTP HELO command filter */
 	mlfi_envfrom,	/* envelope sender filter */
@@ -2493,6 +2665,15 @@ struct smfiDesc smfilter =
 	mlfi_eom,	/* end of message */
 	mlfi_abort,	/* message aborted */
 	mlfi_close,	/* shutdown */
+#if SMFI_VERSION > 2
+	NULL,		/* unrecognised command */
+#endif
+#if SMFI_VERSION > 3
+	NULL,		/* DATA */
+#endif
+#if SMFI_VERSION >= 0x01000000
+	mlfi_negotiate,	/* negotiation callback */
+#endif
 };
 #endif /* !DEBUG */
 
@@ -2849,6 +3030,9 @@ main(int argc, char **argv)
 	peerlist = NULL;
 	internal = NULL;
 	exignore = NULL;
+#ifdef _FFR_QUARANTINE
+	quarantine = FALSE;
+#endif /* _FFR_QUARANTINE */
 	canon = DK_CANON_SIMPLE;
 	tmo = DEFTIMEOUT;
 	if (DK_DEBUG('t'))
@@ -2981,6 +3165,12 @@ main(int argc, char **argv)
 			pidfile = optarg;
 			break;
 
+#if _FFR_QUARANTINE
+		  case 'q':
+			quarantine = TRUE;
+			break;
+#endif /* _FFR_QUARANTINE */
+
 #if _FFR_REQUIRED_HEADERS
 		  case 'r':
 			req_hdrs = TRUE;
@@ -3034,7 +3224,7 @@ main(int argc, char **argv)
 		  case 'V':
 			printf("%s: %s v%s\n", progname, DKF_PRODUCT,
 			       DKF_VERSION);
-			dkf_ffrlist(stdout);
+			dkf_optlist(stdout);
 			return EX_OK;
 
 		  default:
@@ -3964,6 +4154,15 @@ main(int argc, char **argv)
 #if POPAUTH
 	if (popdbfile != NULL)
 	{
+		status = dkf_initpopauth();
+		if (status != 0)
+		{
+			fprintf(stderr, "%s: can't initialize mutex: %s\n",
+			        progname, strerror(status));
+			syslog(LOG_ERR, "can't initialize mutex: %s",
+			       popdbfile);
+		}
+
 		status = 0;
 
 # if DB_VERSION_MAJOR > 2
