@@ -2,18 +2,18 @@
 **  Copyright (c) 2004-2007 Sendmail, Inc. and its suppliers.
 **	All rights reserved.
 **
-**  $Id: dk-filter.h,v 1.88 2007/03/13 21:16:54 msk Exp $
+**  $Id: dk-filter.h,v 1.93 2007/04/10 21:18:56 msk Exp $
 */
 
 #ifndef _DK_FILTER_H_
 #define _DK_FILTER_H_
 
 #ifndef lint
-static char dk_filter_h_id[] = "@(#)$Id: dk-filter.h,v 1.88 2007/03/13 21:16:54 msk Exp $";
+static char dk_filter_h_id[] = "@(#)$Id: dk-filter.h,v 1.93 2007/04/10 21:18:56 msk Exp $";
 #endif /* !lint */
 
 #define	DKF_PRODUCT	"Sendmail DomainKeys Filter"
-#define	DKF_VERSION	"0.4.2"
+#define	DKF_VERSION	"0.5.0"
 
 /* libdk */
 #include <dk.h>
@@ -28,7 +28,7 @@ static char dk_filter_h_id[] = "@(#)$Id: dk-filter.h,v 1.88 2007/03/13 21:16:54
 
 /* defaults, limits, etc. */
 #define	BUFRSZ		256
-#define CMDLINEOPTS	"a:Ab:c:C:d:DfFhHi:I:klm:M:o:p:P:rRs:S:T:u:U:V"
+#define CMDLINEOPTS	"a:Ab:c:C:d:DfFhHi:I:klm:M:o:p:P:qrRs:S:T:u:U:V"
 #define	DEFTIMEOUT	5
 #define	HEADERMARGIN	75
 #define	HOSTUNKNOWN	"unknown-host"
