@@ -1,10 +1,10 @@
 /*
-**  Copyright (c) 2004-2006 Sendmail, Inc. and its suppliers.
+**  Copyright (c) 2004-2007 Sendmail, Inc. and its suppliers.
 **    All rights reserved.
 */
 
 #ifndef lint
-static char ar_c_id[] = "@(#)$Id: ar.c,v 1.60 2006/09/16 23:00:59 msk Exp $";
+static char ar_c_id[] = "@(#)$Id: ar.c,v 1.61 2007/04/04 17:07:14 msk Exp $";
 #endif /* !lint */
 
 /* OS stuff */
@@ -33,6 +33,7 @@ static char ar_c_id[] = "@(#)$Id: ar.c,v 1.60 2006/09/16 23:00:59 msk Exp $";
 #include <unistd.h>
 #include <pthread.h>
 #include <assert.h>
+#include <signal.h>
 
 /* libsm includes */
 #include <sm/gen.h>
@@ -701,6 +702,7 @@ ar_dispatcher(void *tp)
 #endif /* SELECT */
 	struct timeval timeout;
 	struct timeval timeleft;
+	sigset_t set;
 
 	assert(tp != NULL);
 
@@ -710,6 +712,13 @@ ar_dispatcher(void *tp)
 
 	lib->ar_resend = 0;
 
+	/* block signals that should be caught elsewhere */
+	sigemptyset(&set);
+	sigaddset(&set, SIGHUP);
+	sigaddset(&set, SIGTERM);
+	sigaddset(&set, SIGINT);
+	pthread_sigmask(SIG_BLOCK, &set, NULL);
+
 	for (;;)
 	{
 		maxfd = MAX(lib->ar_nsfd, lib->ar_control[1]);
