@@ -2,11 +2,11 @@
 **  Copyright (c) 2004-2007 Sendmail, Inc. and its suppliers.
 **    All rights reserved.
 **
-**  $Id: dk.c,v 1.173 2007/03/13 18:10:53 msk Exp $
+**  $Id: dk.c,v 1.180 2007/04/10 01:42:50 msk Exp $
 */
 
 #ifndef lint
-static char dk_c_id[] = "@(#)$Id: dk.c,v 1.173 2007/03/13 18:10:53 msk Exp $";
+static char dk_c_id[] = "@(#)$Id: dk.c,v 1.180 2007/04/10 01:42:50 msk Exp $";
 #endif /* !lint */
 
 /* system includes */
@@ -75,6 +75,7 @@ static char dk_c_id[] = "@(#)$Id: dk.c,v 1.173 2007/03/13 18:10:53 msk Exp $";
 
 /* limits, macros, etc. */
 #define BUFRSZ			1024	/* block I/O buffer size */
+#define	DEFERRLEN		128	/* default error buffer size */
 #define	DEFTIMEOUT		5	/* default DNS timeout */
 #define MAXPARAMETER		256	/* biggest parameter we accept */
 
@@ -97,6 +98,9 @@ static char dk_c_id[] = "@(#)$Id: dk.c,v 1.173 2007/03/13 18:10:53 msk Exp $";
 #define	DK_PRIVATEKEY		"/var/db/domainkeys/%s.key.pem"
 					/* template filename for private keys */
 
+/* prototypes */
+static void dk_error(DK *dk, const char *format, ...);
+
 /* parameter lookup table */
 struct lookup
 {
@@ -299,6 +303,12 @@ dk_decode(DK *dk, int what)
 	{
 		dk->dk_key = dk_malloc(dk->dk_libhandle, dk->dk_closure,
 		                       dk->dk_b64len);
+		if (dk->dk_key == NULL)
+		{
+			dk_error(dk, "unable to allocate %d byte(s)",
+			         dk->dk_b64len);
+			return FALSE;
+		}
 		memset(dk->dk_key, '\0', dk->dk_b64len);
 		x = dk->dk_key;
 	}
@@ -306,14 +316,17 @@ dk_decode(DK *dk, int what)
 	{
 		dk->dk_signature = dk_malloc(dk->dk_libhandle, dk->dk_closure,
 		                             dk->dk_b64len);
+		if (dk->dk_signature == NULL)
+		{
+			dk_error(dk, "unable to allocate %d byte(s)",
+			         dk->dk_b64len);
+			return FALSE;
+		}
 		memset(dk->dk_signature, '\0', dk->dk_b64len);
 		x = dk->dk_signature;
 		dk->dk_flags |= DK_FLAG_FREESIGNATURE;
 	}
 
-	if (x == NULL)
-		return FALSE;
-
 	keyidx = 0;
 
 	for (c = dk->dk_b64, n = 0;
@@ -765,28 +778,42 @@ dk_get_key(DK *dk)
 
 		ssel = dk_sterilize(dk->dk_selector);
 		if (ssel == NULL)
+		{
+			dk_error(dk, "unsafe selector data");
 			return DK_STAT_SYNTAX;
+		}
 
 		memset(fn, '\0', sizeof fn);
 		snprintf(fn, sizeof fn - 1, DK_PRIVATEKEY,
 		         dk->dk_selector == NULL ? DK_DEFAULT_SELECTOR : ssel);
 		f = fopen(fn, "r");
 		if (f == NULL)
+		{
+			dk_error(dk, "%s: fopen(): %s", fn, strerror(errno));
 			return DK_STAT_INTERNAL;
+		}
 
 		if (fstat(fileno(f), &s) != 0)
+		{
+			dk_error(dk, "%s: fstat(): %s", fn, strerror(errno));
 			return DK_STAT_INTERNAL;
+		}
 
 		dk->dk_key = dk_malloc(dk->dk_libhandle, dk->dk_closure,
 		                       (size_t) s.st_size);
 		if (dk->dk_key == NULL)
+		{
+			dk_error(dk, "unable to allocate %d byte(s)",
+			         s.st_size);
 			return DK_STAT_NORESOURCE;
+		}
 		dk->dk_keylen = s.st_size;
 
 		n = fread(dk->dk_key, (unsigned int) s.st_size, 1, f);
 		if (n != 1)
 		{
 			fclose(f);
+			dk_error(dk, "%s: fread(): %s", fn, strerror(errno));
 			return DK_STAT_INTERNAL;
 		}
 
@@ -829,6 +856,8 @@ dk_get_key(DK *dk)
 			syslog(LOG_NOTICE, "%s: ar_waitreply(): %s", dk->dk_id,
 			       ar_strerror(error));
 		}
+
+		dk_error(dk, "DNS error or timeout for `%s'", qname);
 		return DK_STAT_CANTVRFY;
 	}
 #else /* USE_ARLIB */
@@ -844,11 +873,13 @@ dk_get_key(DK *dk)
 		{
 		  case HOST_NOT_FOUND:
 		  case NO_DATA:
+			dk_error(dk, "`%s' not found", qname);
 			return DK_STAT_NOKEY;
 
 		  case TRY_AGAIN:
 		  case NO_RECOVERY:
 		  default:
+			dk_error(dk, "DNS error or timeout for `%s'", qname);
 			return DK_STAT_CANTVRFY;
 		}
 	}
@@ -870,7 +901,10 @@ dk_get_key(DK *dk)
 		                 sizeof qname);
 
 		if ((n = dn_skipname(cp, eom)) < 0)
+		{
+			dk_error(dk, "key record corrupted");
 			return DK_STAT_INTERNAL;
+		}
 		cp += n;
 
 		/* extract the type and class */
@@ -881,20 +915,32 @@ dk_get_key(DK *dk)
 	}
 
 	if (type != T_TXT || class != C_IN)
+	{
+		dk_error(dk, "unexpected key reply");
 		return DK_STAT_INTERNAL;
+	}
 
 	/* if NXDOMAIN, return DK_STAT_NOKEY */
 	if (hdr.rcode == NXDOMAIN)
+	{
+		dk_error(dk, "`%s' not found", qname);
 		return DK_STAT_NOKEY;
+	}
 
 	/* get the answer count */
 	ancount = ntohs((unsigned short) hdr.ancount);
 	if (ancount == 0)
+	{
+		dk_error(dk, "DNS error or timeout for `%s'", qname);
 		return DK_STAT_CANTVRFY;
+	}
 
 	/* if truncated, we can't do it */
 	if (hdr.tc)
+	{
+		dk_error(dk, "DNS reply for `%s' truncated", qname);
 		return DK_STAT_INTERNAL;
+	}
 
 	/*
 	**  Extract the data from the first answer.
@@ -903,13 +949,19 @@ dk_get_key(DK *dk)
 	/* grab the label, even though we know what we asked... */
 	if ((n = dn_expand((unsigned char *) &ansbuf, eom, cp,
 	                   (RES_UNC_T) qname, sizeof qname)) < 0)
+	{
+		dk_error(dk, "key record corrupted");
 		return DK_STAT_INTERNAL;
+	}
 	/* ...and move past it */
 	cp += n;
 
 	/* extract the type and class */
 	if (cp + INT16SZ + INT16SZ > eom)
+	{
+		dk_error(dk, "key record corrupted");
 		return DK_STAT_INTERNAL;
+	}
 	GETSHORT(type, cp);
 	GETSHORT(class, cp);
 
@@ -918,16 +970,25 @@ dk_get_key(DK *dk)
 
 	/* get payload length */
 	if (cp + INT16SZ > eom)
+	{
+		dk_error(dk, "key record corrupted");
 		return DK_STAT_INTERNAL;
+	}
 	GETSHORT(n, cp);
 
 	/* XXX -- maybe deal with a partial reply rather than require it all */
 	if (cp + n > eom)
-		return -1;
+	{
+		dk_error(dk, "key record corrupted");
+		return DK_STAT_INTERNAL;
+	}
 
 	dk->dk_b64 = dk_malloc(dk->dk_libhandle, dk->dk_closure, n);
 	if (dk->dk_b64 == NULL)
+	{
+		dk_error(dk, "unable to allocate %d byte(s)", n);
 		return DK_STAT_NORESOURCE;
+	}
 	dk->dk_b64len = n;
 
 	/* decode the tags, allocate memory, and copy */
@@ -938,11 +999,20 @@ dk_get_key(DK *dk)
 		if (p == eom || n == 0)
 		{
 			if (state == 3)
+			{
 				state = 4;
+			}
 			else if (state == 1 || state == 2)
+			{
+				dk_error(dk,
+				         "syntax error in key record at `%s'",
+				         qname);
 				return DK_STAT_CANTVRFY;
+			}
 			else
+			{
 				break;
+			}
 		}
 
 		/* if appropriate, consume a length byte */
@@ -1026,7 +1096,12 @@ dk_get_key(DK *dk)
 			{
 				/* key type */
 				if (strcasecmp(value, "rsa") != 0)
+				{
+					dk_error(dk,
+					         "unknown key type `%s' at `%s'",
+					         value, qname);
 					return DK_STAT_CANTVRFY;
+				}
 			}
 			else if (strcasecmp(tag, "t") == 0)
 			{
@@ -1463,7 +1538,8 @@ dk_hdrsigned(DK *dk, u_char *hdr, bool mark)
 
 	for (c = 0; c < dk->dk_hdrlidx; c++)
 	{
-		if (strncasecmp(dk->dk_hdrlist[c], hdr, colon - hdr) == 0)
+		if (strncasecmp(dk->dk_hdrlist[c], hdr, colon - hdr) == 0 &&
+		    strlen(dk->dk_hdrlist[c]) == colon - hdr)
 		{
 			if (mark)
 				dk->dk_hdrmark[c] = TRUE;
@@ -1554,6 +1630,99 @@ dk_canon(DK *dk, unsigned char *buf, size_t buflen)
 	SHA1_Update(&sha1->sha1_sha1, buf, buflen);
 }
 
+/*
+**  DK_VERROR -- log an error into a DK handle (varargs version)
+**
+**  Parameters:
+**  	dk -- DK context in which this is performed
+**  	format -- format to apply
+**  	va -- argument list
+**
+**  Return value:
+**  	None.
+*/
+
+static void
+dk_verror(DK *dk, const char *format, va_list va)
+{
+	int flen;
+	int saverr;
+	char *new;
+
+	assert(dk != NULL);
+	assert(format != NULL);
+
+	saverr = errno;
+
+	if (dk->dk_error == NULL)
+	{
+		dk->dk_error = dk_malloc(dk->dk_libhandle, dk->dk_closure,
+		                         DEFERRLEN);
+		if (dk->dk_error == NULL)
+		{
+			errno = saverr;
+			return;
+		}
+		dk->dk_errlen = DEFERRLEN;
+	}
+
+	for (;;)
+	{
+		flen = vsnprintf(dk->dk_error, dk->dk_errlen, format, va);
+
+		/* compensate for broken vsnprintf() implementations */
+		if (flen == -1)
+			flen = dk->dk_errlen * 2;
+
+		if (flen >= dk->dk_errlen)
+		{
+			new = dk_malloc(dk->dk_libhandle, dk->dk_closure,
+			                flen + 1);
+			if (new == NULL)
+			{
+				errno = saverr;
+				return;
+			}
+
+			dk_mfree(dk->dk_libhandle, dk->dk_closure,
+			         dk->dk_error);
+			dk->dk_error = new;
+			dk->dk_errlen = flen + 1;
+		}
+		else
+		{
+			break;
+		}
+	}
+
+	errno = saverr;
+}
+
+/*
+**  DK_ERROR -- log an error into a DK handle
+**
+**  Parameters:
+**  	dk -- DK context in which this is performed
+**  	format -- format to apply
+**  	... -- arguments
+**
+**  Return value:
+**  	None.
+*/
+
+static void
+dk_error(DK *dk, const char *format, ...)
+{
+	va_list va;
+
+	assert(dk != NULL);
+	assert(format != NULL);
+
+	va_start(va, format);
+	dk_verror(dk, format, va);
+	va_end(va);
+}
+
 /*
 **  ===== PUBLIC SECTION
 */
@@ -1661,6 +1830,19 @@ dk_verify(DK_LIB *dklib, const char *id, void *memclosure, DK_STAT *statp)
 	return new;
 }
 
+/*
+**  DK_GETERROR -- return any stored error string from within the DKIM
+**                 context handle
+*/
+
+const char *
+dk_geterror(DK *dk)
+{
+	assert(dk != NULL);
+
+	return dk->dk_error;
+}
+
 /*
 **  DK_HEADER -- process a header
 */
@@ -1798,7 +1980,10 @@ dk_header(DK *dk, u_char *hdr, size_t len)
 
 		status = rfc2822_mailbox_split(colon + 1, &user, &domain);
 		if (status != 0 || user == NULL || domain == NULL)
+		{
+			dk_error(dk, "unable to parse sender `%s'", colon + 1);
 			return DK_STAT_SYNTAX;
+		}
 		sm_strlcpy(dk->dk_uhdrn, hdr,
 		        MIN(sizeof dk->dk_uhdrn, (u_int) (colon - hdr) + 1));
 		dk_lowercase(dk->dk_uhdrn);
@@ -1822,7 +2007,10 @@ dk_header(DK *dk, u_char *hdr, size_t len)
 			hval = hdr + strlen(DK_SIGNHEADER) + 1;
 
 			if (dk_process_dkheader(dk, hval) != 0)
+			{
+				dk_error(dk, "syntax error in signature");
 				return DK_STAT_SYNTAX;
+			}
 
 			dk->dk_processing = TRUE;
 		}
@@ -1869,7 +2057,10 @@ dk_eoh(DK *dk)
 		if (strncasecmp(dk->dk_hdrset[c], "from:", 5) == 0)
 		{
 			if (from != NULL)
+			{
+				dk_error(dk, "multiple From: headers found");
 				return DK_STAT_SYNTAX;
+			}
 
 			from = dk->dk_hdrset[c];
 		}
@@ -1877,7 +2068,10 @@ dk_eoh(DK *dk)
 		if (strncasecmp(dk->dk_hdrset[c], "sender:", 7) == 0)
 		{
 			if (sender != NULL)
+			{
+				dk_error(dk, "multiple Sender: headers found");
 				return DK_STAT_SYNTAX;
+			}
 
 			sender = dk->dk_hdrset[c];
 		}
@@ -1897,7 +2091,10 @@ dk_eoh(DK *dk)
 
 		at = strchr(dk->dk_uhdrv, '@');
 		if (at == NULL || *(at + 1) == '\0')
+		{
+			dk_error(dk, "missing or empty domain name in sender");
 			return DK_STAT_SYNTAX;
+		}
 
 		/* clean up if dk->dk_domain already allocated */
 		if (dk->dk_domain == NULL)
@@ -1910,7 +2107,11 @@ dk_eoh(DK *dk)
 		                          dk->dk_closure,
 		                          strlen(at + 1) + 1);
 		if (dk->dk_domain == NULL)
+		{
+			dk_error(dk, "unable to allocate %d byte(s)",
+			         strlen(at + 1) + 1);
 			return DK_STAT_NORESOURCE;
+		}
 		sm_strlcpy(dk->dk_domain, at + 1, strlen(at + 1) + 1);
 
 		dk->dk_skipbody = TRUE;
@@ -1939,7 +2140,10 @@ dk_eoh(DK *dk)
 		sm_strlcpy(dk->dk_frombuf, colon + 1, sizeof dk->dk_frombuf);
 		status = rfc2822_mailbox_split(dk->dk_frombuf, &user, &domain);
 		if (status != 0 || user == NULL || domain == NULL)
+		{
+			dk_error(dk, "unable to parse sender `%s'", colon + 1);
 			return DK_STAT_SYNTAX;
+		}
 		snprintf(dk->dk_sender, len, "%s@%s", user, domain);
 
 		/* verify domain/subdomain match */
@@ -1949,7 +2153,10 @@ dk_eoh(DK *dk)
 
 			/* incorrect DomainKey-Signature: header? */
 			if (dk->dk_domain == NULL)
+			{
+				dk_error(dk, "can't determine signing domain");
 				return DK_STAT_CANTVRFY;
+			}
 
 			if (strcasecmp(domain, dk->dk_domain) != 0)
 			{
@@ -1971,7 +2178,11 @@ dk_eoh(DK *dk)
 			}
 
 			if (!ok)
+			{
+				dk_error(dk,
+				         "signing domain and sending domain mismatch");
 				return DK_STAT_CANTVRFY;
+			}
 		}
 
 		dk->dk_user = dk_strdup(dk->dk_libhandle, dk->dk_closure,
@@ -1985,6 +2196,7 @@ dk_eoh(DK *dk)
 	}
 	else
 	{
+		dk_error(dk, "no sender header found");
 		return DK_STAT_SYNTAX;
 	}
 
@@ -1992,7 +2204,11 @@ dk_eoh(DK *dk)
 	sha1 = dk_malloc(dk->dk_libhandle, dk->dk_closure,
 	                 sizeof(struct dk_sha1));
 	if (sha1 == NULL)
+	{
+		dk_error(dk, "unable to allocate %d byte(s)",
+		         sizeof(struct dk_sha1));
 		return DK_STAT_NORESOURCE;
+	}
 	memset(sha1, '\0', sizeof(struct dk_sha1));
 	sha1->sha1_tmpfd = -1;
 	dk->dk_signinfo = (void *) sha1;
@@ -2001,7 +2217,10 @@ dk_eoh(DK *dk)
 	if (dk->dk_gran != NULL && dk->dk_gran[0] != '\0')
 	{
 		if (strcmp(dk->dk_user, dk->dk_gran) != 0)
+		{
+			dk_error(dk, "key granularity mismatch");
 			return DK_STAT_CANTVRFY;
+		}
 	}
 
 	SHA1_Init(&sha1->sha1_sha1);
@@ -2021,6 +2240,8 @@ dk_eoh(DK *dk)
 
 			errno = saveerr;
 
+			dk_error(dk, "unable to create temporary file at %s",
+			         sha1->sha1_tmppath);
 			return DK_STAT_NORESOURCE;
 		}
 		sha1->sha1_tmpbio = BIO_new_fd(sha1->sha1_tmpfd, 1);
@@ -2097,7 +2318,11 @@ dk_eoh(DK *dk)
 			sm_strlcpy(addr, dk->dk_uhdrv, sizeof addr);
 			status = rfc2822_mailbox_split(addr, &user, &domain);
 			if (status != 0 || user == NULL || domain == NULL)
+			{
+				dk_error(dk, "unable to parse sender `%s'",
+				         addr);
 				return DK_STAT_SYNTAX;
+			}
 			snprintf(dk->dk_uhdrv, sizeof dk->dk_uhdrv, "%s@%s",
 			         user, domain);
 			at = strchr(dk->dk_uhdrv, '@');
@@ -2138,13 +2363,21 @@ dk_eoh(DK *dk)
 
 			at = strchr(dk->dk_sender, '@');
 			if (at == NULL || *(at + 1) == '\0')
+			{
+				dk_error(dk, "can't find sender domain in `%s'",
+				         dk->dk_sender);
 				return DK_STAT_SYNTAX;
+			}
 
 			dk->dk_domain = dk_malloc(dk->dk_libhandle,
 			                          dk->dk_closure,
 			                          strlen(at + 1) + 1);
 			if (dk->dk_domain == NULL)
+			{
+				dk_error(dk, "unable to allocate %d byte(s)",
+				         strlen(at + 1) + 1);
 				return DK_STAT_NORESOURCE;
+			}
 			sm_strlcpy(dk->dk_domain, at + 1, strlen(at + 1) + 1);
 		}
 
@@ -2418,7 +2651,10 @@ dk_eom(DK *dk, DK_FLAGS *dkf)
 
 		/* no sender header was found below the signature */
 		if (dk->dk_skipbody && dk->dk_processing)
+		{
+			dk_error(dk, "no sender header found below signature");
 			return DK_STAT_SYNTAX;
+		}
 
 		if (dk->dk_key == NULL)
 		{
@@ -2439,10 +2675,14 @@ dk_eom(DK *dk, DK_FLAGS *dkf)
 		/* load the public key */
 		key = BIO_new_mem_buf(dk->dk_key, dk->dk_keylen);
 		if (key == NULL)
+		{
+			dk_error(dk, "BIO_new_mem_buf() failed");
 			return DK_STAT_NORESOURCE;
+		}
 		sha1->sha1_pkey = d2i_PUBKEY_bio(key, NULL);
 		if (sha1->sha1_pkey == NULL)
 		{
+			dk_error(dk, "d2i_PUBKEY_bio() failed");
 			BIO_free(key);
 			return DK_STAT_NORESOURCE;
 		}
@@ -2451,6 +2691,7 @@ dk_eom(DK *dk, DK_FLAGS *dkf)
 		sha1->sha1_rsa = EVP_PKEY_get1_RSA(sha1->sha1_pkey);
 		if (sha1->sha1_rsa == NULL)
 		{
+			dk_error(dk, "EVP_PKEY_get1_RSA() failed");
 			BIO_free(key);
 			return DK_STAT_NORESOURCE;
 		}
@@ -2500,7 +2741,7 @@ dk_getsig(DK *dk, u_char *buf, size_t len)
 	int c;
 	int char_count;
 	int status;
-	size_t l;
+	u_int l;
 	size_t b64idx;
 	size_t b64len;
 	unsigned char *b64;
@@ -2514,13 +2755,19 @@ dk_getsig(DK *dk, u_char *buf, size_t len)
 
 	/* verify state */
 	if (dk->dk_state != DK_STATE_EOM)
+	{
+		dk_error(dk, "dk_getsig() called out of sequence");
 		return DK_STAT_INTERNAL;
+	}
 
 	sha1 = (struct dk_sha1 *) dk->dk_signinfo;
 
 	/* must be called only with a signing handle */
 	if (dk->dk_mode == DK_MODE_VERIFY)
+	{
+		dk_error(dk, "dk_getsig() called with verifying handle");
 		return DK_STAT_INTERNAL;
+	}
 
 	/* try to get the signature if it wasn't given */
 	if (dk->dk_key == NULL)
@@ -2535,10 +2782,14 @@ dk_getsig(DK *dk, u_char *buf, size_t len)
 	/* load the private key */
 	key = BIO_new_mem_buf(dk->dk_key, dk->dk_keylen);
 	if (key == NULL)
+	{
+		dk_error(dk, "BIO_new_mem_buf() failed");
 		return DK_STAT_NORESOURCE;
+	}
 	sha1->sha1_pkey = PEM_read_bio_PrivateKey(key, NULL, NULL, NULL);
 	if (sha1->sha1_pkey == NULL)
 	{
+		dk_error(dk, "PEM_read_bio_PrivateKey() failed");
 		BIO_free(key);
 		return DK_STAT_NORESOURCE;
 	}
@@ -2547,6 +2798,7 @@ dk_getsig(DK *dk, u_char *buf, size_t len)
 	sha1->sha1_rsa = EVP_PKEY_get1_RSA(sha1->sha1_pkey);
 	if (sha1->sha1_rsa == NULL)
 	{
+		dk_error(dk, "EVP_PKEY_get1_RSA() failed");
 		BIO_free(key);
 		return DK_STAT_NORESOURCE;
 	}
@@ -2557,6 +2809,8 @@ dk_getsig(DK *dk, u_char *buf, size_t len)
 	                              sha1->sha1_keysize);
 	if (sha1->sha1_rsaout == NULL)
 	{
+		dk_error(dk, "unable to allocate %d byte(s)",
+		         sha1->sha1_keysize);
 		BIO_free(key);
 		return DK_STAT_NORESOURCE;
 	}
@@ -2577,6 +2831,7 @@ dk_getsig(DK *dk, u_char *buf, size_t len)
 	if (status == 0 || l == 0)
 	{
 		BIO_free(key);
+		dk_error(dk, "RSA_sign() returned %d, length %d", status, l);
 		return DK_STAT_INTERNAL;
 	}
 
@@ -2592,6 +2847,7 @@ dk_getsig(DK *dk, u_char *buf, size_t len)
 	b64 = dk_malloc(dk->dk_libhandle, dk->dk_closure, b64len);
 	if (b64 == NULL)
 	{
+		dk_error(dk, "unable to allocate %d byte(s)", b64len);
 		BIO_free(key);
 		return DK_STAT_NORESOURCE;
 	}
@@ -2609,6 +2865,7 @@ dk_getsig(DK *dk, u_char *buf, size_t len)
 			if (b64idx > b64len - 4)
 			{
 				BIO_free(key);
+				dk_error(dk, "base64 encoding failed");
 				return DK_STAT_INTERNAL;
 			}
 
@@ -2630,6 +2887,7 @@ dk_getsig(DK *dk, u_char *buf, size_t len)
 		if (b64idx > b64len - 4)
 		{
 			BIO_free(key);
+			dk_error(dk, "base64 encoding failed");
 			return DK_STAT_INTERNAL;
 		}
 
