@@ -7,7 +7,7 @@
 #define _SMI_DK_H_
 
 #ifndef lint
-static char dk_h_id[] = "@(#)$Id: dk.h,v 1.48 2006/06/01 17:55:48 msk Exp $";
+static char dk_h_id[] = "@(#)$Id: dk.h,v 1.49 2007/03/13 23:02:34 msk Exp $";
 #endif /* !lint */
 
 /* system includes */
@@ -349,6 +349,19 @@ extern DK_STAT dk_timeout __P((DK *dk, int new, int *old));
 
 extern DK_STAT dk_free __P((DK *dk));
 
+/*
+**  DK_GETERROR -- return any stored error string from within the DK
+**                 context handle
+**
+**  Parameters:
+**  	dk -- DK handle from which to retrieve an error string
+**
+**  Return value:
+**  	A pointer to the stored string, or NULL if none was stored.
+*/
+
+extern const char *dk_geterror __P((DK *dk));
+
 /*
 **  rfc2822_mailbox_split -- extract the userid and host from a structured
 **                           header
